@@ -32,10 +32,6 @@
 (require 'speechd)
 
 
-(defconst speechd-speak-version "2004-05-27 08:26 pdm"
-  "Version of the speechd-speak file.")
-
-
 ;;; User options
 
 
@@ -101,7 +97,7 @@ See also `speechd-speak-buffer-insertions'."
   :type '(repeat (string :tag "Buffer name"))
   :group 'speechd-speak)
 
-(defcustom speechd-speak-priority-insertions-in-buffers '()
+(defcustom speechd-speak-priority-insertions-in-buffers '("*shell*")
   "List of names of buffers, in which insertions are spoken immediately.
 Unlike `speechd-speak-insertions-in-buffers', speaking is not delayed until a
 command is completed.
@@ -213,12 +209,24 @@ If ACTION is a function, it is invoked, with no arguments."
                                  :match #'(lambda (w val) (commandp val)))))
   :group 'speechd-speak)
 
+(defcustom speechd-speak-display-modes '(gnus-agent-mode)
+  "List of minor modes to be spoken by their display string rather than name."
+  :type '(repeat symbol)
+  :group 'speechd-speak)
+
 (defcustom speechd-speak-whole-line nil
   "If non-nil, speak whole line on movement by default.
-Otherwise from the point to the end of line on movement by default."
+Otherwise speak from the point to the end of line on movement by default."
   :type 'boolean
   :group 'speechd-speak)
 
+(defcustom speechd-speak-message-time-interval 30
+  "Minimum time in seconds, after which the same message may be repeated.
+If the message is the same as the last one, it is not spoken unless the number
+of seconds defined here has passed from the last spoken message."
+  :type 'integer
+  :group 'speechd-speak)
+
 (defcustom speechd-speak-connections '()
   "Alist mapping major modes and buffers to speechd connection.
 By default, there's a single connection to speechd, named \"default\".  This
@@ -365,6 +373,13 @@ Level 1 is the slowest, level 9 is the fastest."
 ;;; Supporting functions and options
 
 
+(defvar speechd-speak-command-done nil
+  "When non-nil, no post command speaking is performed, except for keys.
+This variable is reset to nil before each command in pre-command-hook.")
+
+(defun speechd-speak--name (&rest args)
+  (intern (mapconcat #'symbol-name args "-")))
+
 (defvar speechd-speak-mode nil)   ; forward definition to make everything happy
 
 (defvar speechd-speak--started nil)
@@ -426,9 +441,13 @@ Level 1 is the slowest, level 9 is the fastest."
        (or (eq major-mode 'debugger-mode)
            (and (boundp 'edebug-active) edebug-active))))
 
-(defmacro speechd-speak--maybe-speak (&rest body)
+(defmacro speechd-speak--maybe-speak* (&rest body)
   `(when (and speechd-speak-mode
               (not (speechd-speak--in-debugger)))
+     ,@body))
+
+(defmacro speechd-speak--maybe-speak (&rest body)
+  `(speechd-speak--maybe-speak*
      (let ((speechd-client-name (speechd-speak--connection-name))
            (speechd-language
             (or (and speechd-speak-input-method-languages
@@ -438,15 +457,6 @@ Level 1 is the slowest, level 9 is the fastest."
                 speechd-language)))
        ,@body)))
 
-(defvar speechd-speak--in-async nil)
-
-(defun speechd-speak--async (&rest args)
-  (if speechd-speak--in-async
-      (push args speechd-speak--pending-speeches)
-    (let ((speechd-speak--in-async t))
-      (speechd-speak--read-pending-speeches)
-      (apply (car args) (cdr args)))))
-
 (defmacro speechd-speak--interactive (&rest body)
   `(let ((speechd-speak-mode (or (interactive-p)
                                  (and speechd-speak-mode
@@ -524,7 +534,7 @@ played."
             (speechd-speak--event-mapping 'empty)
             "")
         :priority speechd-default-text-priority))
-      ((string-match "\\`[ \t]+\\'" text)
+      ((save-match-data (string-match "\\`[ \t]+\\'" text))
        (speechd-speak-report (or (speechd-speak--event-mapping 'whitespace) "")
                              :priority speechd-default-text-priority))
       (t
@@ -589,6 +599,14 @@ If BUFFER is nil, read current buffer."
   (speechd-speak--interactive
    (speechd-speak-read-buffer (window-buffer (get-lru-window)))))
 
+(defun speechd-speak-read-mode-line ()
+  "Read mode line.
+This function works only in Emacs 21.4 or higher."
+  (interactive)
+  (when (fboundp 'format-mode-line)
+    (speechd-speak--interactive
+     (speechd-speak--text (format-mode-line)))))
+
 (defun speechd-speak--window-contents ()
   (sit-for 0)                           ; to update window start and end
   (speechd-speak-read-region (window-start) (window-end)))
@@ -608,12 +626,11 @@ If BUFFER is nil, read current buffer."
   (speechd-speak--text (format "Column %d" (current-column))))
 
 (defmacro speechd-speak--def-speak-object (type)
-  (let* ((name (symbol-name type))
-	 (function-name (intern (format "speechd-speak-read-%s" name)))
-	 (backward-function (intern (format "backward-%s" name)))
-	 (forward-function (intern (format "forward-%s" name))))
+  (let* ((function-name (speechd-speak--name 'speechd-speak-read type))
+	 (backward-function (speechd-speak--name 'backward type))
+	 (forward-function (speechd-speak--name 'forward type)))
     `(defun ,function-name ()
-       ,(format "Speak current %s." name)
+       ,(format "Speak current %s." type)
        (interactive)
        (speechd-speak--interactive
         (save-excursion
@@ -640,10 +657,11 @@ If BUFFER is nil, read current buffer."
   other-window
   other-buffer-modified
   completion-buffer-modified
-  minibuffer-contents)
+  minibuffer-contents
+  info)
 
 (defmacro speechd-speak--cinfo (slot)
-  `(,(intern (concat "speechd-speak--command-info-struct-" (symbol-name slot)))
+  `(,(speechd-speak--name 'speechd-speak--command-info-struct slot)
     info))
 
 (defun speechd-speak--command-info-struct-buffer (info)
@@ -694,7 +712,9 @@ If BUFFER is nil, read current buffer."
                  :minibuffer-contents
                    (if (speechd-speak--in-minibuffer-p)
                        (minibuffer-contents)
-                     'unset))))))))
+                     'unset)
+                 :info (speechd-speak--current-info)
+                )))))))
 
 (defun speechd-speak--reset-command-start-info ()
   (speechd-speak--set-command-start-info t))
@@ -751,13 +771,15 @@ FUNCTION is invoked interactively."
 				ad-do-it)
 			   `(when (interactive-p)
 			      ,body*))))
-		 commands*))))
+		 commands*)
+       ,(unless (eq position 'before)
+          `(setq speechd-speak-command-done t)))))
 
 (defmacro speechd-speak--command-feedback-region (commands)
   `(speechd-speak--command-feedback ,commands around
      (let ((start (point)))
-       ad-do-it
-       (speechd-speak--speak-piece start))))
+       (prog1 ad-do-it
+         (speechd-speak--speak-piece start)))))
 
 (defun* speechd-speak--next-property-change (&optional (point (point))
                                                        (limit (point-max)))
@@ -791,22 +813,6 @@ connections, otherwise create completely new connection."
     (set (make-local-variable 'speechd-client-name) name)
     (setq speechd-speak--client-name-set t)))
 
-(defvar speechd-speak--pending-speeches '())
-
-(defvar speechd-speak--reading-pending-speeches nil)
-
-(defun speechd-speak--read-pending-speeches ()
-  (unless (or (speechd-blocked-p)
-              speechd-speak--reading-pending-speeches)
-    (let ((speechd-speak--reading-pending-speeches t))
-      ;; Beware, speechd-speak--pending-speeches can change during the
-      ;; processing!
-      (while speechd-speak--pending-speeches
-        (let ((request (car speechd-speak--pending-speeches)))
-          (apply (car request) (cdr request)))
-        (setq speechd-speak--pending-speeches
-              (cdr speechd-speak--pending-speeches))))))
-
 
 ;;; Basic speaking
 
@@ -852,16 +858,16 @@ connections, otherwise create completely new connection."
 				 around
   (when speechd-speak-deleted-char
     (speechd-speak-read-char (preceding-char)))
-  ad-do-it
-  (unless speechd-speak-deleted-char
-    (speechd-speak-read-char (preceding-char))))
+  (prog1 ad-do-it
+    (unless speechd-speak-deleted-char
+      (speechd-speak-read-char (preceding-char)))))
 
 (speechd-speak--command-feedback (delete-char) around
   (when speechd-speak-deleted-char
     (speechd-speak-read-char (following-char)))
-  ad-do-it
-  (unless speechd-speak-deleted-char
-    (speechd-speak-read-char (following-char))))
+  (prog1 ad-do-it
+    (unless speechd-speak-deleted-char
+      (speechd-speak-read-char (following-char)))))
 
 (speechd-speak--command-feedback (quoted-insert) after
   (speechd-speak-read-char (preceding-char)))
@@ -872,6 +878,15 @@ connections, otherwise create completely new connection."
 (speechd-speak--command-feedback (undo) after
   (speechd-speak-read-line))
 
+(defmacro speechd-speak--unhide-message (function)
+  ;; If `message' is invoked within a built-in function, there's no way to get
+  ;; notified automatically about it.  So we have to wrap the built-in
+  ;; functions displaying messages to check for the otherwise hidden messages.
+  `(speechd-speak--defadvice ,function after
+     (speechd-speak--current-message)))
+
+(speechd-speak--unhide-message write-region)
+
 
 ;;; Killing and yanking
 
@@ -901,9 +916,9 @@ connections, otherwise create completely new connection."
 
 (speechd-speak--command-feedback (kill-region completion-kill-region) around
   (let ((nlines (count-lines (region-beginning) (region-end))))
-    ad-do-it
-    (speechd-speak--maybe-speak
-     (message "Killed region containing %s lines" nlines))))
+    (prog1 ad-do-it
+      (speechd-speak--maybe-speak*
+        (message "Killed region containing %s lines" nlines)))))
 
 
 ;;; Messages
@@ -911,6 +926,7 @@ connections, otherwise create completely new connection."
 
 (defvar speechd-speak--last-message "")
 (defvar speechd-speak--last-spoken-message "")
+(defvar speechd-speak--last-spoken-message-time 0)
 
 (defun speechd-speak-last-message ()
   "Speak last message from the echo area."
@@ -921,18 +937,34 @@ connections, otherwise create completely new connection."
      (speechd-speak--text speechd-speak--last-message))))
 
 (defun speechd-speak--read-message (message)
-  (let ((speechd-speak--special-area t)
-        (speechd-spell nil))
-    (speechd-speak--signal 'message :priority 'progress)
-    (speechd-speak--minibuffer-prompt message :priority 'progress)))
+  (let ((speechd-speak--special-area t))
+    (speechd-speak--maybe-speak         ; necessary to select proper connection
+      (speechd-block '(spelling-mode nil message-priority progress)
+        (speechd-speak--signal 'message)
+        (speechd-speak--minibuffer-prompt message)))))
 
 (defun speechd-speak--message (message &optional reset-last-spoken)
-  (speechd-speak--maybe-speak
+  (speechd-speak--maybe-speak*
    (when (and message
-              (not (string= message speechd-speak--last-spoken-message)))
-     (setq speechd-speak--last-message message
-           speechd-speak--last-spoken-message message)
-     (speechd-speak--async #'speechd-speak--read-message message))
+              (or (not (string= message speechd-speak--last-spoken-message))
+                  (>= (- (float-time) speechd-speak--last-spoken-message-time)
+                      speechd-speak-message-time-interval)))
+     (let* ((oldlen (length speechd-speak--last-spoken-message))
+            (len (length message))
+            ;; The following tries to handle answers to y-or-n questions, e.g.
+            ;; "Do something? (y or n) y", which are reported as a single
+            ;; string, including the prompt.
+            (message* (if (and (= (- len oldlen) 1)
+                               (save-match-data
+                                 (string-match "\?.*) .$" message))
+                               (string= (substring message 0 oldlen)
+                                        speechd-speak--last-spoken-message))
+                          (substring message oldlen)
+                        message)))
+       (setq speechd-speak--last-message message
+             speechd-speak--last-spoken-message message
+             speechd-speak--last-spoken-message-time (float-time))
+       (speechd-speak--read-message message*)))
    (when reset-last-spoken
      (setq speechd-speak--last-spoken-message ""))))
 
@@ -1018,11 +1050,22 @@ This command applies to buffers defined in
 (speechd-speak--command-feedback minibuffer-message after
   (speechd-speak--minibuffer-prompt (ad-get-arg 0) :priority 'notification))
 
-;; The following functions don't invoke `minibuffer-setup-hook'
-(speechd-speak--defadvice y-or-n-p before
-  (let ((speechd-speak--emulate-minibuffer t))
-    (speechd-speak--minibuffer-prompt
-     (concat (ad-get-arg 0) "(y or n)") :priority 'message)))
+;; Some built-in functions, reading a single character answer, prompt in the
+;; echo area.  They don't invoke minibuffer-setup-hook and may put other
+;; messages to the echo area by invoking other built-in functions.  There's no
+;; easy way to catch the prompts and messages, the only way to deal with this
+;; is to use idle timers.
+(defvar speechd-speak--message-timer nil)
+(defun speechd-speak--message-timer ()
+  (let ((message (current-message)))
+    (when (and cursor-in-echo-area
+               (not (string= message speechd-speak--last-spoken-message)))
+      (let ((speechd-speak--emulate-minibuffer t))
+        (speechd-speak--minibuffer-prompt message))
+      (setq speechd-speak--last-spoken-message message))))
+
+;; The following functions don't invoke `minibuffer-setup-hook' and don't put
+;; the cursor into the echo area.  Sigh.
 (speechd-speak--defadvice read-key-sequence before
   (let ((prompt (ad-get-arg 0))
         (speechd-speak--emulate-minibuffer t))
@@ -1036,11 +1079,11 @@ This command applies to buffers defined in
         (speechd-speak--minibuffer-prompt prompt :priority 'message)))))
 
 
-;;; Repitition in character reading commands
+;;; Repetition in character reading commands
 
 
 (defvar speechd-speak-read-char-keymap (make-sparse-keymap)
-  "Keymap used by speechd-el for repititions during reading characters.
+  "Keymap used by speechd-el for repetitions during reading characters.
 Only single characters are allowed in the keymap.")
 (define-key speechd-speak-read-char-keymap
   "\C-a" 'speechd-speak-last-insertions)
@@ -1153,13 +1196,14 @@ Only single characters are allowed in the keymap.")
   (speechd-speak--with-command-start-info
     (unless (= beg end)
       (cond
-       ((or (member (buffer-name)
-                    speechd-speak-priority-insertions-in-buffers)
+       ((or (member (buffer-name) speechd-speak-priority-insertions-in-buffers)
             ;; Asynchronous buffer changes
             (and (not this-command)
                  (member (buffer-name) speechd-speak-insertions-in-buffers)))
-        (speechd-speak--async #'speechd-speak--read-buffer-change (buffer-name)
-                              (speechd-speak--buffer-substring beg end)))
+        (unless (memq this-command '(self-insert-command quoted-insert newline
+                                     newline-and-indent undo yank yank-pop))
+          (speechd-speak--read-buffer-change
+           (buffer-name) (speechd-speak--buffer-substring beg end))))
        ((not this-command)
         ;; Asynchronous buffer change -- we are not interested in it by
         ;; default
@@ -1182,11 +1226,11 @@ Only single characters are allowed in the keymap.")
 (defun speechd-speak--pre-command-hook ()
   (condition-case err
       (progn
-        (speechd-speak--read-pending-speeches)
         (unless (memq this-command speechd-speak--dont-cancel-on-commands)
           (speechd-cancel 1))
         (speechd-speak--set-command-start-info)
         (setq speechd-speak--last-report "")
+        (setq speechd-speak-command-done nil)
         (when speechd-speak-spell-command
           (speechd-speak-spell-mode 1))
         (speechd-speak--maybe-speak
@@ -1211,8 +1255,7 @@ Only single characters are allowed in the keymap.")
   (add-hook 'pre-command-hook 'speechd-speak--pre-command-hook))
 
 (defmacro speechd-speak--post-defun (name shy new-state guard &rest body)
-  (let* ((name (intern (concat "speechd-speak--post-read-"
-                               (symbol-name name))))
+  (let* ((name (speechd-speak--name 'speechd-speak--post-read name))
          (state-condition (case shy
                             ((t) `(eq state nil))
                             (sometimes `(not (eq state t)))
@@ -1228,7 +1271,8 @@ Only single characters are allowed in the keymap.")
 
 (speechd-speak--post-defun special-commands t t
   ;; Speak commands that can't speak in a regular way
-  (memq this-command '(forward-char backward-char))
+  (and (not speechd-speak-command-done)
+       (memq this-command '(forward-char backward-char)))
   (speechd-block `(message-priority ,speechd-default-char-priority)
     (cond
      ((looking-at "^")
@@ -1239,10 +1283,10 @@ Only single characters are allowed in the keymap.")
       (speechd-speak-report (speechd-speak--event-mapping 'end-of-line)))
      (t
       (speechd-speak-read-char)))))
-  
+
 (speechd-speak--post-defun buffer-switch t t
   ;; Any buffer switch
-  buffer-changed
+  (and buffer-changed (not speechd-speak-command-done))
   (when speechd-speak-buffer-name
     (speechd-speak--text (buffer-name) :priority 'message))
   (when (memq speechd-speak-buffer-name '(text nil))
@@ -1263,6 +1307,24 @@ Only single characters are allowed in the keymap.")
       (speechd-speak--text (symbol-name this-command) :priority 'message)
     (speechd-speak--command-keys 'message)))
 
+(speechd-speak--post-defun command-done t t
+  ;; No speaking when the command has already been handled in a special way
+  speechd-speak-command-done)
+
+(speechd-speak--post-defun info-change t nil
+  ;; General status information has changed
+  (not (equal (speechd-speak--cinfo info) (speechd-speak--current-info)))
+  (let ((old-info (speechd-speak--cinfo info))
+        (new-info (speechd-speak--current-info)))
+    (dolist (item new-info)
+      (let* ((id (car item))
+             (new (cdr item))
+             (old (cdr (assq id old-info))))
+        (when (and (memq id speechd-speak-state-changes)
+                   (not (equal old new)))
+          (funcall (speechd-speak--name 'speechd-speak--update id)
+                   old new))))))
+
 (speechd-speak--post-defun speaking-commands nil t
   ;; Avoid additional reading on speaking commands
   (let ((command-name (symbol-name this-command))
@@ -1323,7 +1385,8 @@ Only single characters are allowed in the keymap.")
       (let ((speechd-language "en")
             (speechd-speak-input-method-languages nil))
         (goto-char (point-min))
-        (re-search-forward "\n\n+" nil t)
+        (save-match-data
+          (re-search-forward "\n\n+" nil t))
         (speechd-speak-read-region (point) (point-max) nil))))
 
 (speechd-speak--post-defun special-face-movement sometimes
@@ -1352,7 +1415,14 @@ Only single characters are allowed in the keymap.")
            (memq this-command speechd-speak-by-properties-always)
            (memq (get-char-property (point) 'face)
                  speechd-speak-by-properties-on-movement))
-       (or (get-char-property (point) 'face)
+       (or (let ((properties (text-properties-at (point)))
+                 (look-props '(face font-lock-face mouse-face display help-echo
+                               keymap local-map read-only field))
+                 (marked nil))
+             (while (and look-props (not marked))
+               (setq marked (plist-member properties (car look-props)))
+               (setq look-props (cdr look-props)))
+             marked)
            (overlays-at (point)))
        (let ((position (speechd-speak--cinfo point)))
          (or (> (speechd-speak--previous-property-change
@@ -1401,6 +1471,8 @@ Only single characters are allowed in the keymap.")
   '(speechd-speak--post-read-special-commands
     speechd-speak--post-read-buffer-switch
     speechd-speak--post-read-command-keys
+    speechd-speak--post-read-command-done
+    speechd-speak--post-read-info-change
     speechd-speak--post-read-speaking-commands
     speechd-speak--post-read-buffer-modifications
     speechd-speak--post-read-completions
@@ -1550,6 +1622,218 @@ When the mode is enabled, all spoken text is spelled."
     (setq speechd-speak-spell-command t)))
   
 
+
+;;; Informatory commands
+
+
+(defvar speechd-speak-info-map (make-sparse-keymap))
+
+(defvar speechd-speak--info-updates nil)
+
+(defmacro* speechd-speak--watch (name get-function
+                                 &key on-change info info-string key)
+  `(locally
+     (fset (quote ,(speechd-speak--name 'speechd-speak--get name))
+           ,get-function)
+     ,(when info
+        `(fset (quote ,(speechd-speak--name 'speechd-speak name 'info))
+               #'(lambda (info)
+                   (interactive)
+                   (funcall ,info info))))
+     ,(when info-string
+        `(fset (quote ,(speechd-speak--name 'speechd-speak name 'info))
+               #'(lambda ()
+                   (interactive)
+                   (speechd-speak--text
+                    (format ,info-string
+                            (funcall
+                             (function ,(speechd-speak--name
+                                         'speechd-speak--get name))))))))
+     ,(when (and (or info info-string) key)
+        `(define-key speechd-speak-info-map ,key
+           (quote ,(speechd-speak--name 'speechd-speak name 'info))))
+     ,(when on-change
+        `(defun ,(speechd-speak--name 'speechd-speak--update name) (old new)
+           (speechd-speak--maybe-speak
+            (let ((speechd-default-text-priority 'message))
+              (funcall ,on-change old new)))))
+     ,(when on-change
+        `(add-to-list 'speechd-speak--info-updates (quote ,name)))))
+
+(speechd-speak--watch buffer-name #'buffer-name
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text (format "Buffer %s" new))))
+
+(speechd-speak--watch buffer-identification
+  #'(lambda ()
+      (when (fboundp 'format-mode-line)
+        (let ((ident (format-mode-line mode-line-buffer-identification)))
+          (set-text-properties 0 (length ident) nil ident)
+          ident)))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "New buffer identification: %s" new))))
+
+(speechd-speak--watch buffer-modified #'buffer-modified-p
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (if new "Buffer modified" "No buffer modification"))))
+
+(speechd-speak--watch buffer-read-only #'(lambda () buffer-read-only)
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (if new "Buffer writable" "Buffer read-only"))))
+
+(defun speechd-speak-buffer-info ()
+  "Speak current buffer information."
+  (interactive)
+  (speechd-speak--text
+   (format "Buffer %s, %s %s %s; %s"
+           (speechd-speak--get-buffer-name)
+           (or vc-mode "")
+           (if (speechd-speak--get-buffer-read-only) "read only" "")
+           (if (speechd-speak--get-buffer-modified) "modified" "")
+           (let ((ident (speechd-speak--get-buffer-identification)))
+             (if ident
+                 (format "buffer identification: %s" ident)
+               "")))))
+(define-key speechd-speak-info-map "b" 'speechd-speak-buffer-info)
+
+(speechd-speak--watch frame-name #'(lambda () (frame-parameter nil 'name))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Frame changed to: %s; was: %s" new old))))
+
+(speechd-speak--watch frame-identification
+  #'(lambda ()
+      (when (fboundp 'format-mode-line)
+        (let ((ident (format-mode-line mode-line-frame-identification)))
+          (set-text-properties 0 (length ident) nil ident)
+          ident)))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Frame identification changed to: %s" new))))
+
+(defun speechd-speak-frame-info ()
+  "Speak current frame information."
+  (interactive)
+  (speechd-speak--text
+   (format "Frame: %s; %s"
+           (speechd-speak--get-frame-name)
+           (let ((ident (speechd-speak--get-frame-identification)))
+             (if ident
+                 (format "frame identification: %s" ident)
+               "")))))
+(define-key speechd-speak-info-map "f" 'speechd-speak-frame-info)
+
+(speechd-speak--watch header-line
+  #'(lambda ()
+      (if (fboundp 'format-mode-line)
+          (let ((line (format-mode-line t)))
+            (if (string= line "") "empty" line))
+        "unknown"))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Header line changed to: %s" new)))
+  :info-string "Header line: %s"
+  :key "h")
+
+(speechd-speak--watch major-mode #'(lambda () mode-name)
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Major mode changed to: %s; was: %s" new old))))
+
+(speechd-speak--watch minor-modes
+  #'(lambda ()
+      (loop for mode-spec in minor-mode-alist
+            for mode = (car mode-spec)
+            when (and (boundp mode) (symbol-value mode))
+            collect (if (memq mode speechd-speak-display-modes)
+                        (cadr mode-spec)
+                      mode)))
+  :on-change #'(lambda (old new)
+                 (flet ((set-difference (x y)
+                          (loop for i in x
+                                unless (memq i y)
+                                collect i)))
+                   (let ((disabled (set-difference old new))
+                         (enabled (set-difference new old)))
+                     (when enabled
+                       (speechd-speak--text
+                        (format "Enabled minor modes: %s" enabled)))
+                     (when disabled
+                       (speechd-speak--text
+                        (format "Disabled minor modes: %s" disabled)))))))
+
+(defun speechd-speak-mode-info ()
+  "Speak information about current major and minor modes."
+  (interactive)
+  (speechd-speak--text
+   (format "Major mode: %s; minor modes: %s"
+           (speechd-speak--get-major-mode)
+           (speechd-speak--get-minor-modes))))
+(define-key speechd-speak-info-map "m" 'speechd-speak-mode-info)
+
+(speechd-speak--watch buffer-file-coding
+  #'(lambda () buffer-file-coding-system)
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Buffer file coding changed to: %s; was %s"
+                          new old))))
+
+(speechd-speak--watch terminal-coding #'terminal-coding-system
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Terminal coding changed to: %s; was: %s" new old))))
+
+(defun speechd-speak-coding-info ()
+  "Speak information about current codings."
+  (interactive)
+  (speechd-speak--text
+   (format "Buffer file coding is %s, terminal coding is %s"
+           (speechd-speak--get-buffer-file-coding)
+           (speechd-speak--get-terminal-coding))))
+(define-key speechd-speak-info-map "c" 'speechd-speak-coding-info)
+
+(speechd-speak--watch input-method #'(lambda ()
+                                       (or current-input-method "none"))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Input method changed to: %s; was: %s" new old)))
+  :info-string "Input method %s"
+  :key "i")
+
+(speechd-speak--watch process
+  #'(lambda ()
+      (let ((process (get-buffer-process (current-buffer))))
+        (and process (process-status process))))
+  :on-change #'(lambda (old new)
+                 (speechd-speak--text
+                  (format "Process status changed to: %s; was: %s" new old)))
+  :info-string "Process status: %s"
+  :key "p")
+
+(defcustom speechd-speak-state-changes
+  '(buffer-identification buffer-read-only frame-name frame-identification
+    major-mode minor-modes buffer-file-coding terminal-coding input-method
+    process)
+  "List of identifiers of the Emacs state changes to be automatically reported.
+The following symbols are valid state change identifiers: `buffer-name',
+`buffer-identification', `buffer-modified', `buffer-read-only', `frame-name',
+`frame-identification', `header-line', `major-mode', `minor-modes',
+`buffer-file-coding', `terminal-coding', `input-method', `process'."
+  :type `(set ,@(mapcar #'(lambda (i) `(const ,i))
+                        (reverse speechd-speak--info-updates)))
+  :group 'speechd-speak)
+
+(defun speechd-speak--current-info ()
+  (sort (mapcar #'(lambda (i)
+                    (cons i (funcall
+                             (speechd-speak--name 'speechd-speak--get i))))
+                speechd-speak--info-updates)
+        #'(lambda (x y) (string< (symbol-name (car x))
+                                 (symbol-name (car y))))))
+
 
 ;;; Mode definition
 
@@ -1580,7 +1864,9 @@ When the mode is enabled, all spoken text is spelled."
 (define-key speechd-speak-mode-map ">" 'speechd-speak-read-rest-of-buffer)
 (define-key speechd-speak-mode-map "\C-a" 'speechd-add-connection-settings)
 (define-key speechd-speak-mode-map "\C-c" 'speechd-speak-new-connection)
+(define-key speechd-speak-mode-map "\C-i" speechd-speak-info-map)
 (define-key speechd-speak-mode-map "\C-l" 'speechd-speak-spell)
+(define-key speechd-speak-mode-map "\C-m" 'speechd-speak-read-mode-line)
 (define-key speechd-speak-mode-map "\C-n" 'speechd-speak-read-next-line)
 (define-key speechd-speak-mode-map "\C-p" 'speechd-speak-read-previous-line)
 (define-key speechd-speak-mode-map "\C-r" 'speechd-speak-read-rectangle)
@@ -1624,9 +1910,10 @@ Null prefix argument turns off the mode."
  :group 'speechd-speak)
 
 (defun speechd-speak--shutdown ()
-  (when (speechd-speak--signal 'finish :priority 'important)
-    (sit-for 3))                        ; not to CANCEL prematurely
-  (global-speechd-speak-mode -1))
+  ;; We don't have to call CANCEL here, since Emacs exit is usually called
+  ;; interactively, so it is preceded by the pre-command CANCEL.  Moreover,
+  ;; calling CANCEL here means trouble with stopping the final exit messages.
+  (speechd-speak--signal 'finish :priority 'important))
 
 ;;;###autoload
 (define-minor-mode speechd-speak-mode
@@ -1703,13 +1990,16 @@ starts blocking your Emacs functions."
   (interactive)
   (setq speechd-speak--started nil)
   (ignore-errors (global-speechd-speak-mode -1))
+  (when speechd-speak--message-timer
+    (cancel-timer speechd-speak--message-timer)
+    (setq speechd-speak--message-timer nil))
   (remove-hook 'pre-command-hook 'speechd-speak--pre-command-hook)
   (remove-hook 'post-command-hook 'speechd-speak--post-command-hook)
   (remove-hook 'after-change-functions 'speechd-speak--after-change-hook)
   (remove-hook 'minibuffer-setup-hook 'speechd-speak--minibuffer-setup-hook)
   (remove-hook 'minibuffer-exit-hook 'speechd-speak--minibuffer-exit-hook)
   (remove-hook 'kill-emacs-hook 'speechd-speak--shutdown)
-  (speechd-close)
+  (speechd-close-all)
   (global-speechd-speak-map-mode -1))
 
 ;;;###autoload
@@ -1728,7 +2018,9 @@ With a prefix argument, close all open connections first."
     (global-speechd-speak-mode 1)
     (global-speechd-speak-map-mode 1)
     (speechd-speak--debug 'start)
-    (speechd-speak--signal 'start)))
+    (speechd-speak--signal 'start)
+    (setq speechd-speak--message-timer
+          (run-with-idle-timer 0 t 'speechd-speak--message-timer))))
 
 
 ;;; Announce
@@ -1737,12 +2029,4 @@ With a prefix argument, close all open connections first."
 (provide 'speechd-speak)
 
 
-;; Local variables:
-;; time-stamp-format: "%:y-%02m-%02d %02H:%02M %u"
-;; time-stamp-time-zone: "UTC"
-;; time-stamp-start: "^(defconst speechd.*-version \""
-;; time-stamp-end: "\""
-;; time-stamp-line-limit: 0
-;; End:
-
 ;;; speechd-speak.el ends here
