@@ -34,10 +34,6 @@
 ;; Functions and variables starting with the prefix `speechd--' are considered
 ;; internal and you shouldn't use them outside this file.
 ;;
-;; *Note*: If you call any of the speechd-el functions asynchronously
-;; (typically within a process filter function), you must wrap the whole
-;; asynchronous code by the `speechd-protect' macro.
-;;
 ;; See docstrings and the Texinfo manual for full details.
 
 ;;; Code:
@@ -52,7 +48,7 @@
 (defgroup speechd ()
   "Speech Dispatcher interface.")
 
-(defcustom speechd-host "localhost"
+(defcustom speechd-host (or (getenv "SPEECHD_HOST") "localhost")
   "Name of the default host running speechd to connect to."
   :type 'string
   :group 'speechd)
@@ -66,6 +62,14 @@
   :type 'integer
   :group 'speechd)
 
+(defcustom speechd-spdsend nil
+  "If string, it names the spdsend binary to be used to talk to SSIP.
+If the new version of accept-process-output is available or the value of this
+variable is nil, Emacs talks to SSIP directly."
+  :type '(choice (const :tag "Don't use spdsend" nil)
+                 (string :tag "spdsend binary"))
+  :group 'speechd)
+
 (defcustom speechd-timeout 3
   "Number of seconds to wait for speechd response."
   :type 'integer
@@ -139,7 +143,7 @@ Dispatcher voice.  Gender can be one of the symbols `male', `female',
 Style can be one of the numbers 1, 2, 3 (style values are likely to be changed
 in future).
 
-The message-priority parameter sets priority of amy message of the voice.  Its
+The message-priority parameter sets priority of any message of the voice.  Its
 value is any of the message priority symbols.
 
 See the corresponding speechd-set-* functions for valid values of other
@@ -183,7 +187,7 @@ definition of its parameters is optional."
 		  ,(speechd--generate-customization-options
 		    speechd--capital-character-modes))
             ;;
-	    (cons :tag "Messsage priority" (const :format "" message-priority)
+	    (cons :tag "Message priority" (const :format "" message-priority)
 		  (speechd-priority-tag :value text))
 	    (cons :tag "Output module" (const :format "" output-module)
 		  string))))
@@ -211,58 +215,6 @@ You must reopen the connections to apply the changes to this variable."
              (speechd-reopen)))           
   :group 'speechd)
 
-(defcustom speechd-connection-parameters '()
-  "This variable is obsolete.  Please use `speechd-connection-voices' instead."
-  :get #'(lambda (name)
-           (mapcar
-            #'(lambda (v)
-                (cons (car v)
-                      (let ((result '())
-                            (params (cdr v)))
-                        (while params
-                          (push (cons (car params) (cadr params)) result)
-                          (setq params (nthcdr 2 params)))
-                        (nreverse result))))
-            (symbol-value name)))
-  :set #'(lambda (name value)
-           (let ((real-value
-                  (mapcar
-                   #'(lambda (item)
-                       (cons (car item)
-                             (apply #'nconc
-                                    (mapcar #'(lambda (ival)
-                                                (list (car ival) (cdr ival)))
-                                            (cdr item)))))
-                   value)))
-             (set-default name real-value))
-           (when (fboundp 'speechd-reopen)
-             (speechd-reopen)))
-  :type `(repeat
-          (cons
-	   :tag "Connection"
-	   (choice (string :tag "Connection name")
-		   (const :tag "Default parameters" :value t))
-	   (set
-	    :tag "Parameters"
-	    (cons :tag "Language" (const :format "" language) string)
-	    (cons :tag "Messsage priority"
-		  (const :format "" message-priority)
-		  (speechd-priority-tag :value text))
-	    (cons :tag "Punctuation mode" (const :format "" punctuation-mode)
-		  ,(speechd--generate-customization-options
-		    speechd--punctuation-modes))
-	    (cons :tag "Capital character mode"
-		  (const :format "" capital-character-mode)
-		  ,(speechd--generate-customization-options
-		    speechd--capital-character-modes))
-	    (cons :tag "Voice" (const :format "" voice) string)
-	    (cons :tag "Rate" (const :format "" rate) integer)
-	    (cons :tag "Pitch" (const :format "" pitch) integer)
-	    (cons :tag "Volume" (const :format "" volume) integer)
-	    (cons :tag "Output module" (const :format "" output-module)
-		  string))))
-  :group 'speechd)
-
 (defcustom speechd-face-voices '()
   "Alist mapping faces to voices.
 Each of the alist element is of the form (FACE . VOICE) where FACE is a face
@@ -296,10 +248,6 @@ language.")
 ;;; Internal constants and configuration variables
 
 
-(defconst speechd--el-version "2004-05-25 15:19 pdm"
-  "Version stamp of the source file.
-Useful only for diagnosing problems.")
-
 (defconst speechd--buffer " *speechd*"
   "Name of the buffer associated with speechd connections.")
 
@@ -317,10 +265,10 @@ Useful only for diagnosing problems.")
                                       "en"))
 
 (defvar speechd--default-connection-parameters
-  '(punctuation-mode "some"
-    spelling-mode "off"
-    capital-character-mode "none"
-    voice "male1"
+  '(punctuation-mode some
+    spelling-mode nil
+    capital-character-mode none
+    voice male1
     rate 0
     pitch 0
     volume 100))
@@ -379,7 +327,7 @@ Useful only for diagnosing problems.")
   port
   (failure-p nil)
   process
-  (process-output nil)
+  (process-output "")
   (paused-p nil)
   (in-block nil)
   (parameters ())
@@ -387,8 +335,7 @@ Useful only for diagnosing problems.")
   (last-command nil))
 
 (defstruct speechd--request
-  string
-  (answer-type 'required))
+  string)
 
 (defvar speechd--connections (make-hash-table :test 'equal)
   "Hash table mapping client names to `speechd-connection' instances.")
@@ -435,17 +382,16 @@ Useful only for diagnosing problems.")
     `(let* ((,$parameters ,parameters)
             (,$orig-parameters ()))
        (unwind-protect
-           (progn
+           (let ((,$cparameters (speechd--connection-parameters connection)))
              (while ,$parameters
                (let* ((,$p (first ,$parameters))
                       (,$v (second ,$parameters))
-                      (,$cparameters
-                       (speechd--connection-parameters connection))
                       (,$orig-v (plist-get ,$cparameters ,$p)))
                  (when (and (not (equal ,$v ,$orig-v))
                             (or ,$v
                                 (not (memq ,$p '(language)))))
-                   (when (plist-member ,$cparameters ,$p)
+                   (when (and (plist-member ,$cparameters ,$p)
+                              (not (memq ,$p '(message-priority))))
                      (push (cons ,$p ,$orig-v) ,$orig-parameters))
                    (speechd--set-parameter ,$p ,$v)))
                (setq ,$parameters (nthcdr 2 ,$parameters)))
@@ -459,8 +405,8 @@ Useful only for diagnosing problems.")
             (age (cdr (assoc 'age parameters)))
             (style (cdr (assoc 'style parameters))))
         (format "%s%s%s"
-                (if (eq age 'child) "CHILD_" "")
-                (if (eq gender 'female) "FEMALE" "MALE")
+                (if (eq age 'child) "child_" "")
+                (if (eq gender 'female) "female" "male")
                 (or style "1")))))
 
 (defun speechd--voice-parameters (voice)
@@ -486,14 +432,6 @@ code."
 (defun speechd--current-language ()
   speechd-language)
 
-(defvar speechd--blocked-p nil)
-
-(defun speechd-blocked-p ()
-  "Return non-nil if speaking is currently blocked.
-Speaking is blocked when another speaking command is executed.  That can
-usually happen if your code is invoked asynchronously, typically in timers."
-  speechd--blocked-p)
-
 
 ;;; Process management functions
 
@@ -505,39 +443,96 @@ usually happen if your code is invoked asynchronously, typically in timers."
 	   (let ((speechd-client-name name))
 	     (speechd-open)))))
 
-(defun speechd--process-name ()
-  (process-name (speechd--connection-process (speechd--connection))))
+(defvar speechd--advanced-apo (condition-case _
+                                  (progn
+                                    (accept-process-output nil 0 0 1)
+                                    t)
+                                (error)))
 
-(defun speechd--connection-filter (process string)
-  (speechd--with-current-connection
-    (let ((c (if (eq process (speechd--connection-process connection))
-                 connection
-               (speechd--connection (substring
-                                     (buffer-name (process-buffer process))
-                                     (1+ (length speechd--buffer)))))))
-      (setf (speechd--connection-process-output c)
-            (concat (speechd--connection-process-output c) string)))))
+(defun speechd--call-spdsend (args &optional input)
+  (with-temp-buffer
+    (if input
+        (progn
+          (insert input)
+          (let ((default-process-coding-system
+                  (cons speechd--coding-system speechd--coding-system))
+                (process-coding-system-alist nil)
+                (last-coding-system-used nil))
+            (apply #'call-process-region (point-min) (point-max)
+                   speechd-spdsend t t nil args)))
+      (apply #'call-process speechd-spdsend nil t nil args))
+    (buffer-string)))
 
-(defun speechd--close-process (connection)
-  (let ((process (speechd--connection-process connection)))
-    (ignore-errors
-      (let ((buffer (process-buffer process)))
-	(when buffer
-	  (kill-buffer buffer))))
-    (ignore-errors
+(defun speechd--process-filter (process output)
+  (when speechd--advanced-apo
+    (speechd--with-current-connection
+     (setf (speechd--connection-process-output connection)
+           (concat (speechd--connection-process-output connection) output)))))
+
+(defun speechd--use-spdsend ()
+  (and (not speechd--advanced-apo) speechd-spdsend))
+
+(defun speechd--open-connection (host port)
+  (if (speechd--use-spdsend)
+      (let* ((answer (speechd--call-spdsend
+                      (list "--open" host (format "%d" port))))
+             (alen (and answer (length answer))))
+        (when (and alen (> alen 1))
+          (setq answer (substring answer 0 (1- alen))))
+        answer)
+    (let ((process (open-network-stream "speechd" nil host port)))
       (when process
-	(delete-process process)))
-    (setf (speechd--connection-process connection) nil)))
+        (set-process-coding-system process
+                                   speechd--coding-system
+                                   (if speechd--advanced-apo
+                                       speechd--coding-system
+                                     'raw-text))
+        (if (fboundp 'set-process-query-on-exit-flag)
+            (set-process-query-on-exit-flag process nil)
+          (process-kill-without-query process))
+        (set-process-filter process #'speechd--process-filter))
+      process)))
 
+(defun speechd--close-connection (connection)
+  (let ((process (speechd--connection-process connection)))
+    (when process
+      (if (speechd--use-spdsend)
+          (speechd--call-spdsend (list "--close" process))
+        (delete-process (speechd--connection-process connection)))
+      (setf (speechd--connection-process connection) nil))))
+
+(defun speechd--send-connection (connection command)
+  (let ((process (speechd--connection-process connection)))
+    (when process
+      (condition-case _
+          (if (speechd--use-spdsend)
+              (speechd--call-spdsend (list "--send" process) command)
+            (process-send-string
+             process
+             (if speechd--advanced-apo
+                 command
+               (encode-coding-string command speechd--coding-system)))
+            (when speechd--advanced-apo
+              (while (not (string-match
+                           speechd--end-regexp
+                           (speechd--connection-process-output connection)))
+                (unless (accept-process-output process speechd-timeout nil 1)
+                  (error "Timeout in communication with Speech Dispatcher"))))
+            (prog1 (speechd--connection-process-output connection)
+              (setf (speechd--connection-process-output connection) "")))
+        (error
+         (speechd--permanent-connection-failure connection)
+         (error "Error in communication with Speech Dispatcher"))))))
+  
 ;;;###autoload
 (defun* speechd-open (&optional host port &key quiet force-reopen)
   "Open connection to Speech Dispatcher running on the given host and port.
 If the connection corresponding to the current `speechd-client-name' value
 already exists, close it and reopen again, with the same connection parameters.
 
-The optional arguments HOST and PORT identify the speechd server location
-differing from the values of `speechd-host' and `speechd-port', see
-`open-network-stream' for closer description of those arguments.
+The optional arguments HOST and PORT identify the speechd server location.
+They can override default values stored in the variables `speechd-host' and
+`speechd-port'.
 
 If the key argument QUIET is non-nil, don't report failures and quit silently.
 If the key argument FORCE-REOPEN is non-nil, try to reopen an existent
@@ -549,17 +544,13 @@ Return the opened connection on success, nil otherwise."
     (let ((host (or host speechd-host))
 	  (port (or port speechd-port)))
       (when connection
-	(speechd--close-process connection)
+	(speechd--close-connection connection)
 	(setq host (speechd--connection-host connection)
 	      port (speechd--connection-port connection)))
       (let* ((name speechd-client-name)
              (voice (cdr (assoc name speechd-connection-voices)))
              (default-parameters (append
                                   (speechd--voice-parameters voice)
-                                  (cdr (assoc speechd-client-name
-                                              speechd-connection-parameters))
-                                  (cdr (assoc t
-                                              speechd-connection-parameters))
                                   speechd--default-connection-parameters))
 	     (parameters (cond
                           ((and connection force-reopen)
@@ -572,34 +563,20 @@ Return the opened connection on success, nil otherwise."
                             default-parameters))
                           (t
                            default-parameters)))
-	     (process (when (or
-                             (not connection)
-                             (not (speechd--connection-failure-p connection))
-                             force-reopen)
-                        (condition-case nil
-                            (open-network-stream
-                             "speechd-proc"
-                             (concat speechd--buffer "-" name)
-                             host port)
-                          (error nil)))))
-	(if process
+	     (id (when (or (not connection)
+                           (not (speechd--connection-failure-p connection))
+                           force-reopen)
+                   (condition-case nil
+                       (speechd--open-connection host port)))))
+	(if id
 	    (progn
-	      ;; The process input encoding is set to raw-text and the coding
-	      ;; conversion is performed by hand in `speechd--send-string'.
-	      ;; The reason is that Emacs 21.3 is buggy when performing the
-	      ;; automated conversion on longer strings.
-	      (set-process-coding-system process
-					 speechd--coding-system 'raw-text)
-	      (set-process-filter process 'speechd--connection-filter)
-	      (process-kill-without-query process)
               (setq connection (make-speechd--connection
-                                :name name :host host :port port
-                                :process process :failure-p (not process)))
+                                :name name :host host :port port :process id))
               (puthash name connection speechd--connections))
 	  (unless quiet
             (setq connection nil)
 	    (message "Connection to Speech Dispatcher failed")))
-	(when process
+	(when id
 	  (speechd--set-connection-name name)
           (setq parameters (append parameters
                                    (list 'language speechd--default-language)))
@@ -623,9 +600,13 @@ Return the opened connection on success, nil otherwise."
   (interactive)
   (let ((connection (speechd--connection name nil)))
     (when connection
-      (speechd--close-process connection)
+      (speechd--close-connection connection)
       (remhash name speechd--connections))))
 
+(defun speechd-close-all ()
+  "Close all speechd connections."
+  (speechd--iterate-clients (speechd-close)))
+
 (defun speechd-reopen ()
   "Close and open again all the connections to speechd."
   (interactive)
@@ -640,12 +621,6 @@ Return the opened connection on success, nil otherwise."
     (message (format "%d connections successfully reopened, %d failures"
 		     number-of-connections number-of-failures))))
 
-(defun speechd-running-p ()
-  "Return non-nil, if the current speechd client name process is running."
-  (let* ((connection (speechd--connection))
-	 (process (and connection (speechd--connection-process connection))))
-    (and process (eq (process-status process) 'open))))
-
 
 ;;; Process communication functions
 
@@ -658,80 +633,35 @@ Return the opened connection on success, nil otherwise."
   (format "^[1-2][0-9][0-9] .*%s" speechd--eol))
 
 (defun speechd--permanent-connection-failure (connection)
-  (speechd--close-process connection)
+  (speechd--close-connection connection)
   (setf (speechd--connection-failure-p connection) t
-	(speechd--connection-process-output connection) nil
 	(speechd--connection-paused-p connection) nil
 	(speechd--connection-parameters connection) ()))
 
-(defun speechd--send-string (string)
-  (speechd--with-current-connection
-    (let ((process (speechd--connection-process connection))
-	  (string (encode-coding-string string speechd--coding-system)))
-      (when process
-	(unwind-protect
-	    (process-send-string process string)
-	  (when (not (speechd-running-p))
-            (speechd--permanent-connection-failure connection)))))))
-
-(defun speechd--process-request (request)
-  (speechd--with-current-connection
-   (save-match-data
-     (flet ((read-command-answer ()
-	     (while (let ((output (speechd--connection-process-output
-				   connection)))
-		      (and output
-			   (not (string-match speechd--end-regexp output))))
-	       (let ((output (speechd--connection-process-output connection)))
-		 ;; We have to be very careful with accept-process-output -- it
-		 ;; can invoke another speechd-el functions in various places.
-		 (let ((speechd--blocked-p t))
-                   (accept-process-output
-		     (speechd--connection-process connection)
-		     speechd-timeout))
-		 (when (string= (speechd--connection-process-output connection)
-				output)
-		   (speechd--permanent-connection-failure connection)
-		   (error "Timeout during communication with speechd."))))
-	     ;; Process the answer
-	     (let ((answer (speechd--connection-process-output connection))
-		   (data '())
-		   success
-		   result)
-               (while (and answer
-                           (string-match speechd--result-regexp answer))
-                 (push (match-string 1 answer) data)
-                 (setq answer (substring answer (match-end 0))))
-	       (setq success (and answer
-				  (string-match speechd--success-regexp
-						answer)))
-	       (unless success
-		 (setq data nil))
-	       (setf (speechd--connection-process-output connection) nil)
-	       (setq result (list success
-				  data
-				  (and answer (substring answer 0 3))
-				  (and answer (substring answer 4))))
-	       result)))
-      ;; Read pending answers
-      (speechd--iterate-connections
-        (when (speechd--connection-process-output connection)
-          (read-command-answer)))
-      ;; Send the request
-      (speechd--send-string (speechd--request-string request))
-      (let ((answer-type (speechd--request-answer-type request))
-            (answer nil))
-        (when (and answer-type
-                   (not (speechd--connection-failure-p connection)))
-          (setf (speechd--connection-process-output connection) "")
-          (when answer-type
-            (read-command-answer))))))))
-
 (defun speechd--send-request (request)
   (speechd--with-current-connection
-    (when (speechd-blocked-p)
-      (error "Speech request within speech request processing"))
-    (speechd--process-request request)))
+   (save-match-data
+     (let ((answer (speechd--send-connection
+                    connection (speechd--request-string request))))
+       (let ((data '())
+             success)
+         (while (and answer
+                     (string-match speechd--result-regexp answer))
+           (push (match-string 1 answer) data)
+           (setq answer (substring answer (match-end 0))))
+         (if (or speechd--advanced-apo (speechd--use-spdsend))
+             (progn
+               (setq success (and answer
+                                  (string-match speechd--success-regexp
+                                                answer)))
+               (unless success
+                 (setq data nil))
+               (list success
+                     data
+                     (and success (substring answer 0 3))
+                     (and success (substring answer 4))))
+           '(t nil nil nil)))))))
+
 
 (defconst speechd--block-commands
   '(("speak")
@@ -767,7 +697,7 @@ Return the opened connection on success, nil otherwise."
   (when (first (or (speechd--send-command "SPEAK") '(t)))
     ;; We must be careful here.  There is no answer from SSIP until all data
     ;; including the terminator is sent.  Thus, if we send the data in pieces,
-    ;; there may be noticable delays when waiting for TCP packet
+    ;; there may be noticeable delays when waiting for TCP packet
     ;; acknowledgments.  So we send all data, including the final dot
     ;; terminator, in a single piece.
     (save-match-data
@@ -785,6 +715,9 @@ Return the opened connection on success, nil otherwise."
                      text i))
           (setq text (replace-match ".." nil nil text 3))
           (setq i (1+ (match-end 0))))))
+    ;; We must remove text properties from the string, otherwise Emacs does
+    ;; strange things when recoding non-ASCII characters to UTF-8.
+    (set-text-properties 0 (length text) nil text)
     (unless (first (or (speechd--send-request
                         (make-speechd--request
                          :string (concat text speechd--eol "." speechd--eol)))
@@ -876,7 +809,7 @@ Language must be an RFC 1766 language code, as a string."
          (docstring
           (format "Set %s of the current connection.
 VALUE must be %s.
-If called with a prefix argument, set it for al connections."
+If called with a prefix argument, set it for all connections."
                   (downcase prompt)
                   (cond
                    ((integerp argdesc*)
@@ -939,16 +872,27 @@ If called with a prefix argument, set it for al connections."
                           (connection-parameters
                            (speechd--connection-parameters connection)))
                       (while connection-parameters
-                        (unless (memq (first connection-parameters)
-                                      '(client-name message-priority))
-                          (push (cons (first connection-parameters)
-                                      (second connection-parameters))
-                                voice-parameters))
-                        (setq connection-parameters
-                              (nthcdr 2 connection-parameters)))
+                        (destructuring-bind (p v . next) connection-parameters
+                          (unless (memq p '(client-name message-priority
+                                            spelling-mode))
+                            (push
+                             (cons
+                              (case p
+                                (voice 'name)
+                                (t p))
+                              (case p
+                                (punctuation-mode
+                                 (cdr (assoc
+                                       v speechd--punctuation-modes)))
+                                (capital-character-mode
+                                 (cdr (assoc
+                                       v speechd--capital-character-modes)))
+                                (t v)))
+                             voice-parameters))
+                          (setq connection-parameters next)))
                       (nreverse voice-parameters))))
-             (remove (assoc speechd-client-name speechd-connection-parameters)
-                     speechd-connection-parameters))
+             (remove (assoc speechd-client-name speechd-voices)
+                     speechd-voices))
             ;; the connection
             speechd-connection-voices
             (cons (cons speechd-client-name voice)
@@ -1001,9 +945,9 @@ of the symbols `important', `message', `text', `notification' or
 The key argument SAY-IF-EMPTY is non-nil, TEXT is sent through SSIP even if it
 is empty."
   (interactive "sText: ")
-  (speechd--set-parameter 'message-priority priority)
   (when (or say-if-empty
             (not (string= text "")))
+    (speechd--set-parameter 'message-priority priority)
     (flet ((properties (point)
              (let ((voice (cdr (assq (get-text-property point 'face text)
                                      speechd-face-voices)))
@@ -1029,9 +973,8 @@ is empty."
                           change-point))
                    (substring (substring text beg end)))
               (if properties
-                  (speechd--with-current-connection
-                    (speechd--with-connection-parameters properties
-                      (speechd--send-text substring)))
+                  (speechd--with-connection-parameters properties
+                    (speechd--send-text substring))
                 (speechd--send-text substring))
               (setq beg end))))))))
 
@@ -1160,12 +1103,4 @@ clients."
 (provide 'speechd)
 
 
-;; Local variables:
-;; time-stamp-format: "%:y-%02m-%02d %02H:%02M %u"
-;; time-stamp-time-zone: "UTC"
-;; time-stamp-start: "^(defconst speechd.*-version \""
-;; time-stamp-end: "\""
-;; time-stamp-line-limit: 0
-;; End:
-
 ;;; speechd.el ends here
